# StatisticalGraphics

The idea of the `StatisticalGraphics` package is to create a powerful, yet easy-to-use solution for creating statistical graphics. 
The package uses [`vega`](https://vega.github.io/vega/) (see also [D3](https://d3js.org/)) for producing the final outputs.

# Examples

```julia
julia> using StatisticalGraphics
julia> using InMemoryDatasets
julia> ds = Dataset(x=1:100, y=rand(100), y2=rand(100) .+ 5, group=rand(1:2, 100));
julia> sgplot(ds, [
                    Line(x=:x, y=:y2, y2axis=true, group=:group),
                    Scatter(x=:x, y=:y2, group=:group)
                  ],
                  nominal = [:group],
                  xaxis = Axis(grid=true),
                  yaxis = Axis(grid=true))
```

![output](/assets/visualization.svg)

```julia
julia> ds = Dataset(x=randn(10^7));
julia> sgplot(ds, Histogram(x=:x, color=:steelblue, outlinethickness=0.5, space=0.5),
                  xaxis = Axis(offset=10, domain=false),
                  yaxis = Axis(offset=10, domain=false, grid=true)
               )
```

![hist_ex](/assets/hist_ex.svg)

**Dubai Weather**

```julia
julia> using DLMReader
julia> dubai_weather = filereader("/assets/dubai_weather.csv", dtformat=Dict(1 => ISODateTimeFormat))
julia> sgplot(dubai_weather, [
        Band(x=:date, lower=:min, upper=:max),
        Line(x=:date, y=:min, color="#4682b4", thickness=1),
        Line(x=:date, y=:max, color="#ff7f0e", thickness=0.5),
        Line(x=:date, y=:pressure, color="#2ca02c", y2axis=true, breaks=true),
        Scatter(x=:date, y=:pressure, color="#2ca02c", size=10, y2axis=true)
    ],
    xaxis=Axis(offset=10, type=:date, grid=true, griddash=[1, 1], title="Date"),
    yaxis=Axis(offset=10, grid=true, griddash=[1, 1], title="Temperature(°C)"),
    y2axis=Axis(offset=10, title="Pressure"))
```

![dubai_ex](/assets/dubai.svg)

Using `BoxPlot` to plot monthly temperature (minimum and maximum), and add second axes for plotting weekly pressure.

```julia
modify!(dubai_weather, :date=>byrow(week)=>:Week)
setformat!(dubai_weather, :date=>month)

sgplot(dubai_weather, [
                        BoxPlot(x=[:min, :max], group=:date, outliers=true),
                        BoxPlot(x=:pressure, group=:Week, opacity=0.5, y2axis=true, outliers=true, x2axis=true)
                      ], 
                      xaxis=Axis(title="Month"),
                      yaxis=Axis(title="Temperature"),
                      y2axis=Axis(title = "Pressure", d3format="f"),
                      x2axis=Axis(values=[1;10:10:50]),
                      height=600
      )
```

![dubai_boxplot](/assets/dubai_boxplot.svg)

```julia
julia> bar_ds = Dataset(x = repeat(1:50, outer=50), group = repeat(1:50, inner = 50))
julia> insertcols!(bar_ds, :y => rand(nrow(bar_ds)))
julia> sgplot(bar_ds, Bar(x=2, response=3, group=1), # refer columns by their index
                        groupcolormodel = Dict(:scheme=>"turbo"),
                        xaxis=Axis(show=false),
                        yaxis=Axis(show=false),
                        legend = false)
```

![bar_random](/assets/bar_random.svg)

**unemployment stacked area plot across industries** 


```julia
julia> unemployment = filereader("assets/unemployment_across_industry.csv", dtformat = Dict(2=>ISODateTimeFormat))
julia> sort!(unemployment, :series, rev=true) # keep alphabetical order
julia> modify!(groupby(unemployment, :date), :count=>cumsum=>:cum_sum)
julia> sort!(unemployment, [:date,:cum_sum], rev=[false,true]) # put the larger area at the back

julia> sgplot(unemployment, Band(x=:date, lower=0.0, upper=:cum_sum, group=:series, opacity=1),
                            nominal = [:series],
                            xaxis=Axis(type=:time, nice=false),
                            yaxis=Axis(title=""),
                            groupcolormodel = Dict(:scheme=>"category20b"),
                            )
```

![unemployment](/assets/unemployment.svg)

**[Revenue by Music Format, 1973–2018](https://observablehq.com/@mbostock/revenue-by-music-format-1973-2018)**

```julia
julia> music = filereader("assets/music.csv")
julia> color_ds = filereader("assets/color_ds.csv")
julia> leftjoin!(music, color_ds, on = :Format)# sort data based on d3js example
julia> sort!(music, [:Year, :order], rev = [false, true]) # rev = true for :order to make the color similar to d3js example

julia> inbillion(x) = x/10^9 # make the yaxis values in billion $
julia> setformat!(music, r"Infla" => inbillion)

julia> sgplot(music, [
                Bar(x = :Year, response = r"Infla", group = :Format, grouporder = :data, legend = :music_leg, outlinethickness = 0, space = 0.05)
              ],
              nominal = [:Format],
              groupcolormodel = reverse!(color_ds[:, :Color]),
              yaxis = Axis(title = "Revenue (billion, adj.)", domain = false),
              xaxis = Axis(values = 1975:5:2015),
              legend = Legend(name = :music_leg, rowspace=0, gridalign = :all, columns = 4, orient = :top, values = color_ds[:, :Format]),
              width = 700
              )
```

![music](/assets/music.svg)

# Examples - sgpanel

**Horsepower vs Make vs Cylinders**

```julia
julia> cars = filereader("assets/cars.csv", types = Dict(9=>Date))
julia> make_fmt(x) = split(x)[1]
julia> setformat!(cars, :Name => make_fmt)
julia> sgpanel(cars, :Cylinders, 
                     Bar(response=:Horsepower, x=:Name, stat=:mean, colorresponse=:Horsepower, colorstat=:mean, categoryorder=:hascending, colormodel=[:green, :red], outlinethickness=0.5, space = 0),
                     nominal = [:Cylinders],
                     layout = :row,
                     linkaxis=:y,
                     proportional=true,
                     stepsize=15,
                     xaxis=Axis(title="Make", angle=-90, baseline=:center, align=:right, ticksize=0, domain=false),
                     yaxis=Axis(title="Horsepower", domain=false),
                     order=true,
                     columnspace = 5
              )
```

![car1](assets/car1.svg)

**usage**

```julia
julia> panel_example = Dataset(rand(1:4, 1000, 10), :auto)
julia> sgplot(panel_example, Bar(x=:x1, group=:x2), nominal = [:x2], width = 100, height = 100)
julia> sgplot(panel_example, Bar(x=:x1, group=:x2, groupdisplay = :cluster), nominal = [:x2], width = 100, height = 100)
julia> sgplot(panel_example, Bar(x=:x1, group=:x2, groupdisplay = :step), nominal = [:x2], width = 100, height = 100)

julia> sgpanel(panel_example, [:x3, :x4], Bar(x=:x1, group=:x2), nominal = [:x2], width = 100, height = 100)
julia> sgpanel(panel_example, [:x3, :x4], Bar(x=:x1, group=:x2, groupdisplay = :step), nominal = [:x2], width = 100, height = 100)
julia> sgpanel(panel_example, [:x3, :x4], Bar(x=:x1, group=:x2), nominal = [:x2], layout = :lattice, width = 100, height = 100)
```

![bar1](assets/bar1.svg)
![bar2](assets/bar2.svg)
![bar3](assets/bar3.svg)
![barpanel1](assets/barpanel1.svg)
![barpanel2](assets/barpanel2.svg)
![barlattice](assets/barlattice.svg)

