function val_opts(opts)
    d = Dict{Symbol,Any}()
    for (opt_name, opt_val) in opts
        d[opt_name] = opt_val
    end
    return d
end

function update_default_opts!(default, vals)
    for (opt_name, opt_val) in vals
        default[opt_name] = opt_val
    end
    default
end

# show an html file
function launch_browser(tmppath::String)
    if Sys.isapple()
        run(`open $tmppath`)
    elseif Sys.iswindows()
        run(`cmd /c start $tmppath`)
    elseif Sys.islinux()
        run(`xdg-open $tmppath`)
    end
end

function _colname_as_string(ds, col)
    if col != 0
        first(names(ds, IMD.index(ds)[col]))
    else
        ""
    end
end

# creates a new scale for each single plot
function addto_color_scale!(vspec, source, name, col, isnominal; color_model = nothing)
    new_scale = Dict{Symbol,Any}()
    new_scale[:domain] = Dict{Symbol,Any}()
    new_scale[:domain][:fields] = Dict{Symbol,Any}[]
    push!(new_scale[:domain][:fields], Dict{Symbol,Any}(:data => source, :field => col))
    # we support two types :point and :linear
    new_scale[:type] = isnominal ? "point" : "linear"
    new_scale[:range] = color_model !== nothing ? color_model : isnominal ? "category" : "diverging"
    new_scale[:name] = name
    push!(vspec[:scales], new_scale)
end

function addto_symbol_scale!(vspec, source, name, col)
    new_scale = Dict{Symbol,Any}()
    new_scale[:domain] = Dict{Symbol,Any}()
    new_scale[:domain][:fields] = Dict{Symbol,Any}[]
    push!(new_scale[:domain][:fields], Dict{Symbol,Any}(:data => source, :field => col))
    #symbol only supports "ordinal" scale
    new_scale[:type] = "ordinal"
    new_scale[:range] = "symbol"
    new_scale[:name] = name
    push!(vspec[:scales], new_scale)
end

function addto_group_scale!(group_scale, source, col, all_args)
    if haskey(group_scale, :type) # it has been initialised
        push!(group_scale[:domain][:fields], Dict{Symbol,Any}(:data => source, :field => col))
    else # we should initial the scale first
        group_scale[:type] = col in all_args.nominal ? "ordinal" : "linear"
        group_scale[:domain] = Dict{Symbol,Any}()
        group_scale[:domain][:fields] = Dict{Symbol,Any}[]
        push!(group_scale[:domain][:fields], Dict{Symbol,Any}(:data => source, :field => col))
        group_scale[:range] = all_args.opts[:groupcolormodel]
    end
end

# the following function is to convert some specific types before sending to js
_convert_values_for_js(x::TimeType) = datetime2unix(DateTime(x)) * 1000 # JS cannot read Julia datetime
_convert_values_for_js(x::Bool) = x ? 1 : 0 # filewrite writes bools as 0/1
_convert_values_for_js(x) = x

function addto_scale!(all_args, which_scale, ds, col)
    if all_args.scale_type[which_scale] === nothing
        if col in all_args.nominal
            all_args.scale_type[which_scale] = :point
        else
            all_args.scale_type[which_scale] = all_args.axes[which_scale].opts[:type]
        end
    else
        # check the consitency of scale type
        if col in all_args.nominal
            if !(all_args.scale_type[which_scale] in [:band, :point])
                throw(ArgumentError("discrete and continuous axis cannot be mixed"))
            end
        end
    end

    if all_args.mapformats
        _f = getformat(ds, col)
    else
        _f = identity
    end

    _fun_ = _convert_values_for_js ∘ _f # we should this one for finding domains

    #TODO we should have option about how to handle missings
    if all_args.axes[which_scale].opts[:order] == :ascending
        _temp_fun = x -> sort(filter(!ismissing, _fun_.(unique(_fun_, x))))
    elseif all_args.axes[which_scale].opts[:order] == :descending
        _temp_fun = x -> sort(filter(!ismissing, _fun_.(unique(_fun_, x))), rev=true)
    else
        _temp_fun = x -> filter(!ismissing, _fun_.(unique(_fun_, x)))
    end

    if all_args.uniscale_col === nothing
        if all_args.scale_type[which_scale] in [:band, :point]
            append!(all_args.scale_ds[which_scale], combine(ds, col => (_temp_fun) => :__scale_col__), promote=true, cols=:union)
        else
            append!(all_args.scale_ds[which_scale], combine(ds, col => (x -> [IMD.minimum(_fun_, x), IMD.maximum(_fun_, x)]) => :__scale_col__), promote=true, cols=:union)
        end
    else
        if all_args.scale_type[which_scale] in [:band, :point]
            append!(all_args.scale_ds[which_scale], combine(gatherby(ds, all_args.uniscale_col, mapformats = all_args.mapformats), col => (_temp_fun) => :__scale_col__), promote=true, cols=:union)
        else
            append!(all_args.scale_ds[which_scale], combine(gatherby(ds, all_args.uniscale_col, mapformats = all_args.mapformats), col => (x -> [IMD.minimum(_fun_, x), IMD.maximum(_fun_, x)]) => :__scale_col__), promote=true, cols=:union)
        end
    end


    # if haskey(in_scale, :domain)
    #     push!(in_scale[:domain][:fields], Dict{Symbol,Any}(:data => source, :field => col))
    # else
    #     in_scale[:type] = isnominal ? "point" : axis.opts[:type]
    #     if axis.opts[:type] in [:time]
    #         delete!(in_scale, :zero)
    #     end
    #     in_scale[:domain] = Dict{Symbol,Any}()
    #     in_scale[:domain][:sort] = isnominal ? true : false
    #     in_scale[:domain][:fields] = Dict{Symbol,Any}[]
    #     push!(in_scale[:domain][:fields], Dict{Symbol,Any}(:data => source, :field => col))
    #     in_scale[:nice] = axis.opts[:nice]
    #     in_scale[:reverse] = axis.opts[:reverse]
    # end
end
function addto_axis!(in_axis, axis, title)
    if !haskey(in_axis, :domain)
        if !axis.opts[:show]
            in_axis[:labels] = false
        end
        in_axis[:domain] = axis.opts[:show] ? axis.opts[:domain] : false
        in_axis[:grid] = axis.opts[:grid]
        in_axis[:gridWidth] = axis.opts[:gridthickness]
        in_axis[:ticks] = axis.opts[:show] ? axis.opts[:ticks] : false
        in_axis[:tickSize] = axis.opts[:ticksize]
        if in_axis[:title] === nothing && axis.opts[:show]
            in_axis[:title] = title
        end
        in_axis[:offset] = axis.opts[:offset]
        in_axis[:labelAngle] = axis.opts[:angle]
        if axis.opts[:baseline] !== nothing
            in_axis[:labelBaseline] = axis.opts[:baseline]
        end
        if axis.opts[:align] !== nothing
            in_axis[:labelAlign] = axis.opts[:align]
        end
        in_axis[:labelOverlap] = axis.opts[:labeloverlap]
        in_axis[:domainColor] = axis.opts[:color]
        in_axis[:gridDash] = axis.opts[:griddash]
        in_axis[:gridColor] = axis.opts[:gridcolor]
        if axis.opts[:tickcount] !== nothing
            in_axis[:tickCount] = axis.opts[:tickcount]
        else
            signal_text = contains(in_axis[:scale], "y") ? "ceil(height/40)" : "ceil(width/40)"
            in_axis[:tickCount] = Dict{Symbol,Any}(:signal => signal_text)
        end
        if axis.opts[:d3format] !== nothing
            in_axis[:format] = axis.opts[:d3format]
        end
        if axis.opts[:values] !== nothing
            in_axis[:values] = axis.opts[:values]
        end

    end
end

# apply default values for legend option in the case that user does not have any preference
function _build_legen!(out_leg, leg_opts, _symbol, _title, _id; opts...)
    out_leg[:name] = _id
    if leg_opts[:title] === nothing
        out_leg[:title] = _title
    else
        out_leg[:title] = leg_opts[:title]
    end
    if leg_opts[:symbol] === nothing 
        if  _symbol !== nothing
            out_leg[:symbolType] = _symbol
        end
    else
        out_leg[:symbolType] = leg_opts[:symbol]
    end
    out_leg[:orient] = leg_opts[:orient]
    out_leg[:columns] = leg_opts[:columns]
    out_leg[:direction] = leg_opts[:direction]
    out_leg[:symbolSize] = leg_opts[:size]
    for opt in opts
        push!(out_leg, opt)
    end
    out_leg[:gridAlign] = leg_opts[:gridalign]
    out_leg[:rowPadding] = leg_opts[:rowspace]
    out_leg[:columnPadding] = leg_opts[:columnspace]
    if leg_opts[:values] !== nothing && leg_opts[:values] isa AbstractVector
        out_leg[:values] = leg_opts[:values]
    end

end

function _fill_scales!(vspec, all_args)
    independent_axes = all_args.independent_axes
    shared_axes = setdiff([1, 2, 3, 4], independent_axes)
    for i in 1:4
        if all_args.scale_type[i] !== nothing
            in_scale = vspec[:scales][i]
            in_scale[:type] = all_args.scale_type[i]
            # some marks like Histogram, set :zero to true
            if !(in_scale[:type] in [:time, :band, :point]) && !haskey(in_scale, :zero)
                in_scale[:zero] = false
            end
            if !(in_scale[:type] in [:point, :band])
                in_scale[:nice] = all_args.axes[i].opts[:nice]
            end
            in_scale[:reverse] = all_args.axes[i].opts[:reverse]
        end
    end
    for i in shared_axes
        in_scale = vspec[:scales][i]
        if !isempty(all_args.scale_ds[i]) && all_args.scale_type[i] !== nothing
            if in_scale[:type] in [:band, :point]
                all_args.scale_ds[i] = combine(all_args.scale_ds[i], :__scale_col__ => (x -> unique(vcat(x...))) => :__scale_col__)
                in_scale[:domain] = all_args.scale_ds[i][:, :__scale_col__][1]
                # fix the issue when domain is singlton
                if !(in_scale[:domain] isa Vector)
                    in_scale[:domain] = [in_scale[:domain]]
                end
            else
                all_args.scale_ds[i] = combine(all_args.scale_ds[i], :__scale_col__ => (x -> [[minimum(vcat(x...)), maximum(vcat(x...))]]) => :__scale_col__)
                in_scale[:domain] = combine(all_args.scale_ds[i], :__scale_col__ => (x -> [minimum(vcat(x...)), maximum(vcat(x...))]) => :__scale_col__)[:, :__scale_col__]
            end
        end
    end
    if !isempty(independent_axes)
        uniscale_col = all_args.uniscale_col
    end
    for i in independent_axes
        in_scale = vspec[:scales][i]
        if !isempty(all_args.scale_ds[i]) && all_args.scale_type[i] !== nothing
            if in_scale[:type] in [:band, :point]
                all_args.scale_ds[i] = combine(gatherby(all_args.scale_ds[i], uniscale_col, mapformats = all_args.mapformats), :__scale_col__ => (x -> unique(vcat(x...))) => :__scale_col__)
            else
                all_args.scale_ds[i] = in_scale[:domain] = combine(gatherby(all_args.scale_ds[i], uniscale_col, mapformats = all_args.mapformats), :__scale_col__ => (x -> [[minimum(x), maximum(x)]]) => :__scale_col__)
            end
            setformat!(all_args.scale_ds[i], uniscale_col => getformat(all_args.ds, uniscale_col))
        end
    end
end

function _crossprod(ds1, ds2)
    res = repeat(ds1, outer=nrow(ds2))
    IMD.hcat!(res, repeat(ds2, inner=nrow(ds1)))
    res
end


function _find_timetype_cols(ds, mapformats)
    if !mapformats
        names(ds, TimeType)
    else
        res = String[]
        for i in 1:ncol(ds)
            if Core.Compiler.return_type(getformat(ds, i), Tuple{eltype(IMD._columns(ds)[i])}) <: Union{Missing,TimeType}
                push!(res, names(ds)[i])
            end
        end
        res
    end
end

function add_dummy_col!(all_args)
    for i in 1:4
        if !isempty(all_args.scale_ds[i])
            insertcols!(all_args.scale_ds[i], 1, :__dummy_column_for_join__ => true)
        end
    end
end

function join_scale_info!(ds, all_args)
    for i in 1:4
        if !isempty(all_args.scale_ds[i])
            leftjoin!(ds, all_args.scale_ds[i], on=intersect(names(ds), names(all_args.scale_ds[i])), mapformats = all_args.mapformats)
            rename!(ds, :__scale_col__ => Symbol("__scale__$(i)__"))
            # we should make sure domain is passed as a vector
            modify!(ds, Symbol("__scale__$(i)__") => byrow(x -> x isa Vector ? x : [x]))

        end
    end
    # for proportional scales we support step size, at this point we set them as false
    insertcols!(ds, :x_step => false)
    insertcols!(ds, :y_step => false)
end

function add_height_width_x_y!(panel_info, all_args)
    if all_args.opts[:layout] == :column
        column_add_height_width_x_y!(panel_info, all_args)
    elseif all_args.opts[:layout] == :row
        row_add_height_width_x_y!(panel_info, all_args)
    elseif all_args.opts[:layout] == :lattice
        lattice_add_height_width_x_y!(panel_info, all_args)
    elseif all_args.opts[:layout] == :panel
        panel_add_height_width_x_y!(panel_info, all_args)
    end
end

function column_add_height_width_x_y!(panel_info, all_args)
    insertcols!(panel_info, :x => 0)
    insertcols!(panel_info, :width => all_args.opts[:width])
    
    ## only row heights can be proportional
    _find_height_proportion!(panel_info, all_args)
    insertcols!(panel_info, :y => panel_info[:, :height] .+ all_args.opts[:rowspace])
    modify!(panel_info, :y => cumsum, :y => (x->lag(x,1, default=0)))
    insertcols!(panel_info, :xaxis => false)
    panel_info[nrow(panel_info), :xaxis] = true
    insertcols!(panel_info, :x2axis => false)
    panel_info[1, :x2axis] = true
    insertcols!(panel_info, :yaxis => true)
    insertcols!(panel_info, :y2axis => true)
    # if header are requested we create their values
    if all_args.opts[:showheaders]
        insertcols!(panel_info, :__row__title__ => getindex.(panel_info[:, :__title_info_for_each_panel__], Symbol(all_args.panelby[1])))
    else
        insertcols!(panel_info, :__row__title__ => missing)
    end
    insertcols!(panel_info, :__column__title__ => missing)

end
function row_add_height_width_x_y!(panel_info, all_args)
    insertcols!(panel_info, :y => 0)
    _find_width_proportion!(panel_info, all_args)
    insertcols!(panel_info, :height => all_args.opts[:height])
    insertcols!(panel_info, :x => panel_info[:, :width] .+ all_args.opts[:columnspace])
    modify!(panel_info, :x => cumsum, :x => (x->lag(x,1, default=0)))
    insertcols!(panel_info, :yaxis => false)
    panel_info[1, :yaxis] = true
    insertcols!(panel_info, :y2axis => false)
    panel_info[nrow(panel_info), :y2axis] = true
    insertcols!(panel_info, :xaxis => true)
    insertcols!(panel_info, :x2axis => true)

    if all_args.opts[:showheaders]
        insertcols!(panel_info, :__column__title__ => getindex.(panel_info[:, :__title_info_for_each_panel__], Symbol(all_args.panelby[1])))
    else
        insertcols!(panel_info, :__column__title__ => missing)
    end
    insertcols!(panel_info, :__row__title__ => missing)
end

function _bool_first_true(x)
    res = zeros(Bool, length(x))
    res[1] = true
    res
end
function _bool_last_true(x)
    res = zeros(Bool, length(x))
    res[end] = true
    res
end

function lattice_add_height_width_x_y!(panel_info, all_args)
   
    # insertcols!(panel_info, :width => all_args.opts[:width])
    # insertcols!(panel_info, :height => all_args.opts[:height])
    _find_width_proportion!(panel_info, all_args)
    _find_height_proportion!(panel_info, all_args)
    insertcols!(panel_info, :y => panel_info[:, :height] .+ all_args.opts[:rowspace])
    insertcols!(panel_info, :x => panel_info[:, :width] .+ all_args.opts[:columnspace])
    modify!(groupby(panel_info, 3, mapformats = all_args.mapformats, threads = false), :x => cumsum, :x =>lag)
    modify!(groupby(panel_info, 2, mapformats = all_args.mapformats, threads = false), :y => cumsum, :y => lag)
    map!(panel_info, x->ismissing(x) ? 0 : x, [:x, :y])
    modify!(groupby(panel_info, 3, mapformats = all_args.mapformats, threads = false), 1 => _bool_first_true => :yaxis)
    modify!(groupby(panel_info, 3, mapformats = all_args.mapformats, threads = false), 1 => _bool_last_true => :y2axis)
    modify!(groupby(panel_info, 2, mapformats = all_args.mapformats, threads = false), 1 => _bool_first_true => :x2axis)
    modify!(groupby(panel_info, 2, mapformats = all_args.mapformats, threads = false), 1 => _bool_last_true => :xaxis)

    if all_args.opts[:showheaders] && all_args.opts[:layout] == :lattice # for :panel we like to produce missing values
        insertcols!(panel_info, :__column__title__ => getindex.(panel_info[:, :__title_info_for_each_panel__], Symbol(all_args.panelby[1])))
    else
        insertcols!(panel_info, :__column__title__ => missing)
    end
    # if there is only one panelby column we still like to have __column__title__ column in the output data set
    if length(all_args.panelby)>1
        if all_args.opts[:showheaders] && all_args.opts[:layout] == :lattice
            insertcols!(panel_info, :__row__title__ => getindex.(panel_info[:, :__title_info_for_each_panel__], Symbol(all_args.panelby[2])))
        else
            insertcols!(panel_info, :__row__title__ => missing)
        end
    else
        insertcols!(panel_info, :__row__title__ => missing)
    end

end
function panel_add_height_width_x_y!(panel_info, all_args)
    if all_args.opts[:rows] === nothing
        number_of_columns = all_args.opts[:columns]
        number_of_rows = Int(ceil(nrow(panel_info) / number_of_columns))
    else
        number_of_rows = all_args.opts[:rows]
        number_of_columns = Int(ceil(nrow(panel_info) / number_of_rows))
    end
    p2 = repeat(1:number_of_rows, inner = number_of_columns)
    p1 = repeat(1:number_of_columns, outer = number_of_rows)
    insertcols!(panel_info, 2, :__panel_new_group_colid__ => p1[1:nrow(panel_info)])
    insertcols!(panel_info, 3, :__panel_new_group_rowid__ => p2[1:nrow(panel_info)])
    lattice_add_height_width_x_y!(panel_info, all_args)
  
end

function _modify_scales_for_panel(scales, info)
    for i in 1:4
        if haskey(scales[i], :type)
            scales[i][:domain] = info[Symbol("__scale__$(i)__")]
            if i < 3
                if (scales[i][:type] in [:point, :band, :discrete]) && info[:x_step]
                    scales[i][:range] = Dict{Symbol, Any}(:step => info[:x_step_size])
                else
                    scales[i][:range] = [0, info[:width]]
                end
            else
                if (scales[i][:type] in [:point, :band, :discrete]) &&  info[:y_step]
                    scales[i][:range] = Dict{Symbol,Any}(:step => info[:y_step_size])
                else
                    scales[i][:range] = [info[:height], 0]
                end
            end
        end
    end
    scales
end
function _modify_axes_for_panel(all_args, axes, info)
    lattice_type = all_args.opts[:layout] != :panel # for lattice type layouts we remove axes titles and add them for all colum and rows
    x1loc = findfirst(x -> x[:scale] == "x1", axes)
    if x1loc !== nothing
        if !info[:xaxis]
            delete!(axes[x1loc], :title)
            axes[x1loc][:ticks] = false
            axes[x1loc][:domain] = false
            axes[x1loc][:labels] = false
        end
        if lattice_type
            if !ismissing(info[:__column__title__]) && info[:xaxis]
                _add_title_for_lattice!(axes[x1loc], info, all_args, true)
            else
                delete!(axes[x1loc], :title)
            end
        end
        if all_args.axes[1].opts[:tickcount] === nothing
            axes[x1loc][:tickCount] = ceil(info[:width] / 40)
        end
    end
    x2loc = findfirst(x -> x[:scale] == "x2", axes)
    if x2loc !== nothing
        if !info[:x2axis]
            delete!(axes[x2loc], :title)
            axes[x2loc][:ticks] = false
            axes[x2loc][:domain] = false
            axes[x2loc][:labels] = false
        end
        if lattice_type
            if !ismissing(info[:__column__title__]) && info[:x2axis]
                _add_title_for_lattice!(axes[x2loc], info, all_args, true)
            else
                delete!(axes[x1loc], :title)
            end
        end
        if all_args.axes[2].opts[:tickcount] === nothing
            axes[x2loc][:tickCount] = ceil(info[:width] / 40)
        end
    end
    y1loc = findfirst(x -> x[:scale] == "y1", axes)
    if y1loc !== nothing
        if !info[:yaxis]
            delete!(axes[y1loc], :title)
            axes[y1loc][:ticks] = false
            axes[y1loc][:domain] = false
            axes[y1loc][:labels] = false
        end
        if lattice_type
            if !ismissing(info[:__row__title__]) && info[:yaxis]
                _add_title_for_lattice!(axes[y1loc], info, all_args, false)
            else
                delete!(axes[y1loc], :title)
            end
        end
        if all_args.axes[3].opts[:tickcount] === nothing
            axes[y1loc][:tickCount] = ceil(info[:height] / 40)
        end
    end
    y2loc = findfirst(x -> x[:scale] == "y2", axes)
    if y2loc !== nothing
        if !info[:y2axis]
            delete!(axes[y2loc], :title)
            axes[y2loc][:ticks] = false
            axes[y2loc][:domain] = false
            axes[y2loc][:labels] = false
        end
        if lattice_type
            if !ismissing(info[:__row__title__]) && info[:y2axis]
                _add_title_for_lattice!(axes[y2loc], info, all_args, false)
            else
                delete!(axes[y2loc], :title)
            end
        end
        if all_args.axes[4].opts[:tickcount] === nothing
            axes[y2loc][:tickCount] = ceil(info[:height] / 40)
        end
    end
    # gridScale must be specified for each cell
    if x1loc !== nothing
        if y1loc !== nothing
            axes[x1loc][:gridScale] = "y1"
        else
            axes[x1loc][:gridScale] = "y2"
        end
    else
        if y1loc !== nothing
            axes[x2loc][:gridScale] = "y1"
        else
            axes[x2loc][:gridScale] = "y2"
        end
    end
    if y1loc !== nothing
        if x1loc !== nothing
            axes[y1loc][:gridScale] = "x1"
        else
            axes[y1loc][:gridScale] = "x2"
        end
    else
        if x1loc !== nothing
            axes[y2loc][:gridScale] = "x1"
        else
            axes[y2loc][:gridScale] = "x2"
        end
    end
    axes
end

function add_filters_to_panel_info_fun(val...; colname, f)
    expr  = ""
    for i in eachindex(colname)
        expr *= " datum['$(colname[i])'] == '$(_convert_values_for_js(f[i](val[i])))' "
        if i != lastindex(colname)
            expr *= " && "
        end
    end
    expr
end


function add_filters!(panel_info, all_args)
    _f = all_args.mapformats ? getformat.(Ref(all_args.ds), all_args.panelby) : repeat([identity], length(all_args.panelby))
    modify!(panel_info, (all_args.panelby...,) => byrow((x...) -> add_filters_to_panel_info_fun(x...; colname=all_args.panelby, f = _f )) => :__filtering_formula__)
end

function add_title_to_panel(val...; colname, f)
    keys = (Symbol.(colname)...,)
    vals = ([f[i](val[i]) for i in 1:length(f)]...,)
    (; zip(keys, vals)...)
end

function add_title_panel!(panel_info, all_args)
    _f = all_args.mapformats ? getformat.(Ref(all_args.ds), all_args.panelby) : repeat([identity], length(all_args.panelby))
    modify!(panel_info, (all_args.panelby...,) => byrow((x...) -> add_title_to_panel(x...; colname=all_args.panelby, f=_f)) => :__title_info_for_each_panel__)
end

function _modify_data_for_panel!(vspec, marks, info, idx)
    for i in eachindex(marks)
        needchange = true
        if haskey(marks[i], :from)
            old_name = marks[i][:from][:facet][:data]
            new_name = old_name * "_$idx"
            marks[i][:from][:facet][:data] = new_name
        elseif haskey(marks[i], :marks)
            old_name = marks[i][:marks][1][:from][:data]
            new_name = old_name * "_$idx"
            marks[i][:marks][1][:from][:data] = new_name
        else
            needchange = false
        end
        if needchange
            push!(vspec[:data], Dict{Symbol, Any}(:name => new_name, :source => old_name, :transform => Dict{Symbol, Any}[Dict{Symbol, Any}(:type=>:filter, :expr => info[:__filtering_formula__])]))
        end
    end
end

function _find_height_proportion!(panel_info, all_args)
    if all_args.opts[:proportional] && all_args.opts[:linkaxis] == :x
        #check we both scale 3 and 4 are discrete
        !all(all_args.scale_type[3:4] .∈ Ref([nothing, :band, :point, :discrete])) && throw(ArgumentError("propertional keyword argument only supported for discrete type axes"))
        type3 = all_args.scale_type[3]
        type4 = all_args.scale_type[4]
        # band has half of step on each side of axis
        max3 = type3 === nothing ? 0 : type3 == :point ? max(1, maximum(length, panel_info[:, :__scale__3__]) - 1) : maximum(length, panel_info[:, :__scale__3__])
        max4 = type4 === nothing ? 0 : type4 == :point ? max(1, maximum(length, panel_info[:, :__scale__4__]) - 1) : maximum(length, panel_info[:, :__scale__4__])
        max_length = max(max3, max4)
    
        # if stepsize is not given we should calculate it
        if all_args.opts[:stepsize] === nothing
            stepsize = Int(ceil(all_args.opts[:height] / max_length))
        else
            stepsize = all_args.opts[:stepsize]
        end
        map!(panel_info, x->true, :y_step)
        insertcols!(panel_info, :y_step_size => stepsize)
        # we should treat point and band differently

        # modify!(panel_info, colname .=> byrow(length) .=> ["____temp4353465464_$i" for i in colname], ["____temp4353465464_$i" for i in colname] => byrow(max) => :height_step, [:height_step, :y_step_size] => byrow(prod) => :height)
        if type3 !== nothing
            modify!(panel_info, :__scale__3__ => byrow(x->type3 == :point ? max(length(x)-1, 1) : length(x))=> "____temp4353465464_3")
        end
        if type4 !== nothing
            modify!(panel_info, :__scale__4__ => byrow(x -> type4 == :point ? max(length(x) - 1, 1) : length(x)) => "____temp4353465464_4")
        end
        modify!(panel_info, r"____temp4353465464_"  => byrow(maximum) => :height_step, [:height_step, :y_step_size] => byrow(prod) => :height)
        select!(panel_info, Not(r"____temp4353465464_"))

    
    else
        insertcols!(panel_info, :height => all_args.opts[:height])
        # map!(panel_info,x->false, :y_step)
    end
end
function _find_width_proportion!(panel_info, all_args)
    if all_args.opts[:proportional] && all_args.opts[:linkaxis] == :y
        #check we both scale 3 and 4 are discrete
        !all(all_args.scale_type[1:2] .∈ Ref([nothing, :band, :point, :discrete])) && throw(ArgumentError("propertional keyword argument only supported for discrete type axes"))
        type1 = all_args.scale_type[1]
        type2 = all_args.scale_type[2]
        # band has half of step on each side of axis
        max1 = type1 === nothing ? 0 : type1 == :point ? max(1, maximum(length, panel_info[:, :__scale__1__]) - 1) : maximum(length, panel_info[:, :__scale__1__])
        max2 = type2 === nothing ? 0 : type2 == :point ? max(1, maximum(length, panel_info[:, :__scale__2__]) - 1) : maximum(length, panel_info[:, :__scale__2__])
        max_length = max(max1, max2)
        # if stepsize is not given we should calculate it
        if all_args.opts[:stepsize] === nothing
            stepsize = Int(ceil(all_args.opts[:width] / max_length))
        else
            stepsize = all_args.opts[:stepsize]
        end
        map!(panel_info,x->true, :x_step)
        insertcols!(panel_info, :x_step_size => stepsize)
        # modify!(panel_info, colname .=> byrow(length) .=> ["____temp4353465464_$i" for i in colname], ["____temp4353465464_$i" for i in colname] => byrow(max) => :width_step, [:width_step, :x_step_size] => byrow(prod) => :width)
        if type1 !== nothing
            modify!(panel_info, :__scale__1__ => byrow(x -> type1 == :point ? max(length(x) - 1, 1) : length(x)) => "____temp4353465464_1")
        end
        if type2 !== nothing
            modify!(panel_info, :__scale__2__ => byrow(x -> type2 == :point ? max(length(x) - 1, 1) : length(x)) => "____temp4353465464_2")
        end
        modify!(panel_info, r"____temp4353465464_" => byrow(maximum) => :width_step, [:width_step, :x_step_size] => byrow(prod) => :width)
        select!(panel_info, Not(r"____temp4353465464_"))
    
    else
        insertcols!(panel_info, :width => all_args.opts[:width])
        # map!(panel_info, x->false, :x_step)
    end
end

function _add_title_for_panel!(newmark, info, all_args, all_axes)
    yaxis_exist = any(x->x[:scale]=="y1", all_axes)
    y2axis_exist = any(x->x[:scale]=="y2", all_axes)
    newmark[:title][:text] = string(info[:__title_info_for_each_panel__])
    newmark[:title][:orient] = "top"
    newmark[:title][:frame] = "group"
    newmark[:title][:limit] = info[:width]
    newmark[:title][:fontSize] = all_args.opts[:headersize]
    # adjust the title location when the current panel has visible axes
    if (info[:yaxis] && yaxis_exist)  || (info[:y2axis] && y2axis_exist)
        newmark[:title][:dy] = ceil(all_args.opts[:headersize]/2)
    end
    newmark[:title][:fontWeight] = all_args.opts[:headerfontweight]
    newmark[:title][:fontStyle] = all_args.opts[:headeritalic] ? "italic" : "normal"
end
function _add_title_for_lattice!(axis, info, all_args, iscolumn)
    axis[:encode] = Dict{Symbol, Any}()
    axis[:encode][:title] = Dict{Symbol, Any}()
    axis[:encode][:title][:update] = Dict{Symbol, Any}()
    n_title = axis[:encode][:title][:update]

    n_title[:text] = Dict{Symbol, Any}(:value => iscolumn ? string(info[:__column__title__]) : string(info[:__row__title__]))
    n_title[:limit] = Dict{Symbol, Any}(:value => iscolumn ? info[:width] : info[:height])
    n_title[:fontSize] = Dict{Symbol, Any}(:value => all_args.opts[:headersize])
    n_title[:fontWeight] = Dict{Symbol, Any}(:value => all_args.opts[:headerfontweight])
    n_title[:fontStyle] = Dict{Symbol, Any}(:value => all_args.opts[:headeritalic] ? "italic" : "normal")
end

# we cannot use title marks for this, because we need to be able to add title to more than one orient
# so we create scale and axes for the outer group mark
function _add_axes_title_for_lattice!(inmk, panel_info, vspec, all_args)
    arg_max_width = argmax(panel_info[:, :x])
    _width = panel_info[arg_max_width, :x] + panel_info[arg_max_width, :width]
    arg_max_height = argmax(panel_info[:, :y])
    _height = panel_info[arg_max_height, :y] + panel_info[arg_max_height, :height]

    inmk[:scales] = Dict{Symbol, Any}[]
    inmk[:axes] = Dict{Symbol, Any}[]

    _titles = Any[nothing, nothing, nothing, nothing]
    for ax in vspec[:axes]
        if haskey(ax, :scale) && haskey(ax, :title)
            _titles[findfirst(isequal(ax[:scale]), ["x1", "x2", "y1", "y2"])] = ax[:title]
        end
    end

            
    push!(inmk[:scales], Dict{Symbol, Any}(:name => "xax", :range => [0, _width]))
    push!(inmk[:scales], Dict{Symbol, Any}(:name => "yax", :range => [_height, 0]))
    if _titles[2] !== nothing
        push!(inmk[:axes], Dict{Symbol, Any}(:orient => "top", :scale => "xax", :labels => false, :ticks => false, :domain => false, :offset => all_args.opts[:axistitleoffset], :title => _titles[2] ))
    end
    if _titles[1] !== nothing
        push!(inmk[:axes], Dict{Symbol, Any}(:orient => "bottom", :scale => "xax", :labels => false, :ticks => false, :domain => false, :offset => _height + all_args.opts[:axistitleoffset], :title => _titles[1] ))
    end
    if _titles[4] !== nothing
        push!(inmk[:axes], Dict{Symbol, Any}(:orient => "right", :scale => "yax", :labels => false, :ticks => false, :domain => false, :offset => _width + all_args.opts[:axistitleoffset], :title => _titles[4] ))
    end
    if _titles[3] !== nothing
        push!(inmk[:axes], Dict{Symbol, Any}(:orient => "left", :scale => "yax", :labels => false, :ticks => false, :domain => false, :offset => all_args.opts[:axistitleoffset], :title => _titles[3]))
    end
end

